package ru.landgrafhomyak.serdha.api.v0.queries

import ru.landgrafhomyak.serdha.api.v0.misc.Column
import ru.landgrafhomyak.serdha.api.v0.misc.IntermediateColumn
import ru.landgrafhomyak.serdha.api.v0.misc.DatabaseType
import ru.landgrafhomyak.serdha.api.v0.misc.Expression
import ru.landgrafhomyak.serdha.api.v0.misc.SelectedTable

public interface SelectQuery<QueryUserExtension : Any> : _Selectable<QueryUserExtension> {
	public val userExtension: QueryUserExtension

	public enum class Order {
		ASC, DESC
	}

	@Suppress("ClassName")
	public interface _CommonSelectCreatorScope<QueryUserExtension : Any> {
		public fun where(expression: IntermediateColumn<Boolean, DatabaseType.BOOLEAN, QueryUserExtension>)
		public fun orderBy(order: Order, vararg expression: IntermediateColumn<*, *, QueryUserExtension>)

		public var limit: UInt

		public var offset: ULong

		public fun <RT, DT : DatabaseType<RT>> returnColumn(expr: IntermediateColumn<RT, DT, QueryUserExtension>): Column<RT, DT, QueryUserExtension>

		public fun <ExpressionUserExtension : Any> mapColumns(
			expression: Expression<ExpressionUserExtension>,
			input: Expression.InputLinker<ExpressionUserExtension, QueryUserExtension>,
			output: Expression.OutputLinker<ExpressionUserExtension, QueryUserExtension>
		)
	}

	public interface JoinCreator<QueryUserExtension : Any, LeftTableUserExtension : Any, RightTableUserExtension : Any> : _CommonSelectCreatorScope<QueryUserExtension> {
		public fun createSelectWithJoin(creator: Scope<QueryUserExtension, LeftTableUserExtension, RightTableUserExtension>): QueryUserExtension

		public interface Scope<QueryUserExtension : Any, LeftTableUserExtension : Any, RightTableUserExtension : Any> {
			public val leftTable: SelectedTable<LeftTableUserExtension, QueryUserExtension>
			public val rightTable: SelectedTable<RightTableUserExtension, QueryUserExtension>
		}
	}

	public interface Creator<QueryUserExtension : Any, TableUserExtension : Any> {
		public fun createSelect(creator: Scope<QueryUserExtension, TableUserExtension>): QueryUserExtension

		public interface Scope<QueryUserExtension : Any, TableUserExtension : Any> : _CommonSelectCreatorScope<QueryUserExtension> {
			public val selectedTable: TableUserExtension
		}
	}
}