package ru.landgrafhomyak.serdha.api.v0.table

import kotlin.jvm.JvmName
import ru.landgrafhomyak.serdha.api.v0.misc.Column
import ru.landgrafhomyak.serdha.api.v0.misc.Expression
import ru.landgrafhomyak.serdha.api.v0.misc.DatabaseType

/**
 * Descriptor of 'check' constraint on columns. Used for schema manipulations.
 *
 * @param TableUserExtension Type of [owner table's][CheckConstraint.table] user expression for static reporting errors when this descriptor passed to wrong table.
 */
public interface CheckConstraint<TableUserExtension : Any> {
	/**
	 * Name of constraint for debugging, errors and raw schema access.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("name")
	public val name: String

	/**
	 * Table that contains this constraint. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("table")
	public val table: Table<TableUserExtension, *>


	/**
	 * Expression used to check data.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("expression")
	public val expression: Expression<Boolean, DatabaseType.BOOLEAN, TableUserExtension>


	/**
	 * Set of columns, checked by this constraint. Calculated from [CheckConstraint.expression]. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("affectedColumns")
	public val affectedColumns: List<Column<*, *, TableUserExtension>>
}